dofile "utils.lua"
dofile "parser.lua"
dofile "library.lua"
--======================================================================================
function move_node (obj, target)
	if obj.parent then
		remove_node (obj)
	end
	obj.parent = target
	if not target.items then
		target.items = {}
	end
	table.insert (target.items, obj)
end


function remove_node (obj)
	local items = obj.parent.items
	for i, item in ipairs(items) do
		if items [i] == obj then
			table.remove (items, i)
			break
		end
	end
	
	if obj.parents then
		local ps
		ps, obj.parents = obj.parents, nil
		while table.getn (ps) ~= 0 do
			obj.parent = ps [table.getn(ps)]
			table.remove (ps)
			remove_node (obj)
		end
	end
end


function show_tree (node, level)
	if not node then
		node, level = world, 0
	end
	
	print (string.rep ("\t", level)..node.uid)
	
	if not node.items then
		return
	else	
		for i, v in ipairs (node.items) do
			show_tree (v, level + 1)
		end
	end
end


function show_parents (node)
	print (node.uid)
	if node.parent then
		show_parents (node.parent)
	end
end
--======================================================================================
ifobject = {}

function ifobject:add (...)
	if not self.items then
		self.items = {}
	end

	if type (arg [1]) == "table" then
		for i, v in ipairs (arg) do
			if not v.parent  then
				move_node (v, self)
			else
				table.insert (self.items, v)
				if not v.parents then
					v.parents = {}
				end
				table.insert (v.parents, self)
			end
		end
	else
		if not self.hooks then
			self.hooks = {}
		end
		table.insert (self.hooks, arg [1])
	end
end


function ifobject:moveto (target)
	move_node (self, target)
end


function ifobject:remove (object)
	if not object then
		if self.parent then
			remove_node (self)
		end
	else
		for i, v in ipairs( self.items ) do
			if v == object then
				table.remove (self.items, i)
				return
			end
		end
	end
end


function ifobject:show ()
	show_tree (self, 0)
end

local ifobject_meta = {
	__index = function( self, key )
		local result
		if self.data.parent then
			local parent = _G[self.data.parent]
			if parent and parent ~= self and key ~= "items" then
				result = parent[key]
			end
		end
		if type(result) == "nil" then
			result = self.data[key]
		end		
		return result
	end,
	__newindex = function( self, key, value )
		if type(self.data[key]) ~= "nil" then
			self.data[key] = value
		else
			rawset(self, key, value)
		end
	end
}

function create_ifobject (uid, name)
	local obj = clone (ifobject, {["uid"] = uid, ["name"] = {}, data = {}, caption = ""})
	setmetatable( obj, ifobject_meta )
	getfenv(0)[uid] = obj
	obj.name = name	
	return obj
end


function ifobject:create (uid, name)
	self:add (create_ifobject (uid, name))
end


function ifobject:find (name, list)	
	if not list then list = {} end

	if string.len (self.name) ~= 0 then
		local addthis = true
		for i, noun in ipairs (name) do
			if not string.find (self.name, noun) and noun ~= "all" then
				addthis = false
				break
			end
		end
		
		if table.getn (name) == 1 and name [1] == "all" then
			addthis = false
		end
		
		if addthis then
			table.insert (list, self)
		end
	end
	
	if self.items then
		for i, item in ipairs (self.items) do
			item:find (name, list)
		end
	end
	
	return list
end

function ifobject:isme( name )
	local rest = self.name
	for i, n in ipairs(name) do rest = string.gsub( rest, n, "" ) end
	return not string.find( rest, "%S" )
end


function ifobject:contains (object)
	if self == object then
		return true
	end
	
	if self.items then
		for i, item in ipairs (self.items) do
			if item:contains (object) then
				return true
			end
		end
	end
	
	return false
end


function ifobject:clone (object, uid)
	local temp1, temp2, temp3
	object.items, object.parent, object.parents, temp1, temp2, temp3 = {}, nil, nil, object.items, object.parent, object.parents
	local result = clone (object)
	object.items, object.parent, object.parents = temp1, temp2, temp3
	result.parent = nil
	result.uid = uid
	getfenv(0) [uid] = result
	result:moveto (self)
end

function ifobject:check (parent, cookie)
	cookie = cookie or { }
	if self ~= _G[self.uid] then
		print ("  "..self.uid.."  "..parent.uid)
		parent:remove (self)
		return
	end

	if cookie[self] then
		print (self.uid.."    "..parent.uid..".  ...")
		parent:remove (self)
		return
	end
	cookie[self] = true
	if self.items then
		for i, v in ipairs( self.items ) do
			v:check (self, cookie)
		end
	end
	cookie[self] = nil
end


function ifobject:belongto (object)
	if self.parent then
		local next = self.parent
		while next do
			if next == object then
				return true
			end
			next = next.parent
		end
	else
		return false
	end
	return false
end


function ifobject:count ()
	if not self.items then
		return 0
	end

	return table.getn (self.items)
end


--======================================================================================
function start ()
	while true do
		io.output():write ('> ')
		parse (io.input():read())
		print ("")
	end
end


function save_node (node, file)
	if node.data then
		file:write (node.uid..".data = ")
		dump_table (node.data, file)
		file:write ("\n")
	end
	
	file:write (node.uid..".name = '"..node.name.."'\n")
	if (node.parent) then
		file:write (node.uid..":moveto ("..node.parent.uid..")\n")
	end
	
	if node.items then
		for i, v in ipairs( node.items ) do
			save_node (node.items [i], file)
		end
	end
end


function save (filename)
	local file = io.open (filename, "w")
	save_node (world, file)
	file:close()
end


function free_node (node)
	while node.items and table.getn (node.items) ~= 0 do
		free_node (node.items [1])
	end
	
	node:remove ()
end


function load (filename)
	free_node (world)

	dofile (filename)
end

--======================================================================================
PARSE_ERROR = {[[
      .]], [[
    .]], [[
,     .]], [[
   "".]], [[
    .]], [[
    .]], [[
     .  .]], [[
,      .]], [[
 - .]]}

sixdays_print_error = print

function parsing_error (error)
	local pe = PARSE_ERROR [error + 1]
	local temp, n = string.gsub (pe, "[^\n\r]+", " ")
	n = math.random (n)
	temp = 1
	string.gsub (pe, "[^\n\r]+", function (s)
		if temp == n then
			sixdays_print_error (s)
		end
		temp = temp + 1
	end)
end
--======================================================================================

create_ifobject ("world", "")
world:create ("player", "")